#pragma once
#define GLM_ENABLE_EXPERIMENTAL
#include <glm/glm.hpp>
#include <glm/gtx/transform.hpp>
#include <glad/glad.h>
#include <iostream>
#include <vector>
#include <algorithm>

#include "camera.h"
#include "shader.h"

static const unsigned int maxParticles = 70000;
static const float GRAVITY = 9.81f;

struct Particle {
	enum Type { LAUNCHING, SPARKLE, TRAIL, DEAD };

	glm::vec3 pos, speed;
	unsigned char r, g, b, a;
	float size, life, trailTime, cameraDst;
	Type type;

	bool operator<(const Particle& right) const {
		return this->cameraDst > right.cameraDst;
	}
};

class Launcher
{
public:
	Launcher();
	Launcher(glm::vec3 position);

	void RenderTrails(Particle& p, float deltaTime);
	void SpawnParticle(glm::vec3 position, glm::vec3 speed, glm::vec4 color, float size, float life, Particle::Type type);
	void Explode(Particle& p);
	void LaunchFirework();

	void Simulate(float dt);
	void Update(float dt);

	void SortParticles();
	int FindUnusedParticle();

	void Init(Camera* camera, glm::mat4& projection);
	void Draw(bool reflect);

	bool Launch = false;


	static unsigned int particlesCount;

private:
	float launchDelay = 0.15f;       /* delay between each launch (seconds) */
	float trailDelay = 0.04f;        /* delay between each trail spawn (seconds) */
	float fountainDelay = 0.05f;

	int sparklesPerExplosion = 300; /* Number of sparkles after explosion */
	float sparkleLife = 0.9f;       /* lifetime of a sparkle (seconds) */
	float sparkleSize = 1.3f;

	float trailSize = 1.6f;         /* size of a trail */
	float trailLife = 0.6f;         /* lifetime of a trail (seconds) */

	float rocketSize = 0.6f;        /* size of a rocket */
	float rocketLife = 1.5f;        /* lifetime of a rocket before exploding (seconds) 8.5f */


	/* Colors */
	glm::vec3 trailFade = glm::vec3(0.0f, 0.0f, 0.0f);

	/* Spread radius */
	float explosionSize = 30.0f;    /* size of explosion blow */
	float explosionSpread = 10.0f;   /* spread of explosion blow */
	float launchSpread = 40.0f;     /* spread of launch direction */
	float launchSpeed = 160.0f;      /* speed of launch direction */

	/* Do not touch */
	std::unique_ptr<Particle[]> particles{ new Particle[maxParticles] };
	glm::vec3 position;
	int lastUsedId = 0;
	float launchTime = 0.0f, fountainTime = 0.0f;
	float trailTime = 0.0f;


	Shader* shader;
	Camera* camera;
	GLfloat* particle_position;
	GLubyte* particle_color;
	GLuint textureId;
	GLuint VAO;
	GLuint billboard_vertex_buffer;
	GLuint particles_position_buffer;
	GLuint particles_color_buffer;
};
