#include <glad/glad.h>
#include <GLFW/glfw3.h>
#include <stb_image.h>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>


#include "camera.h"
#include "model.h"

#include <iostream>
#include "util.h"
#include "light.h"
#include "skybox.h"
#include "framebuffer.h"
#include "shadow.h"
#include "launcher.h"

#define SHADER "../resource/shaders/"
#define TEX "../resource/textures/"
#define OBJ "../resource/objects/"

void framebuffer_size_callback(GLFWwindow* window, int width, int height);
void processInput(GLFWwindow* window);
void mouse_callback(GLFWwindow* window, double xpos, double ypos);
void mouse_button_callback(GLFWwindow* window, int button, int action, int mods);
void scroll_callback(GLFWwindow* window, double xoffset, double yoffset);

// settings
const unsigned int SCR_WIDTH = 1920;
const unsigned int SCR_HEIGHT = 1080;



// Matrix
glm::mat4 projection;
glm::mat4 view;

// timing
float deltaTime = 0.0f;
float lastFrame = 0.0f;

GLuint cubeVAO;
GLuint quadVAO;

#pragma region models
Model* group;
Mesh* fluid;
Mesh* terrain;
Model* forest;
Model* houses;

Shader* staticShader;
Shader* fluidShader;
unsigned int fluidEffect = 1;
void initModels();
void renderModels(glm::mat4 view, bool inv, bool isDrawFluid, bool isDrawDepth);
#pragma endregion


#pragma region framebuffer
Framebuffer* fbRefraction;
Framebuffer* fbReflection;
Framebuffer* fb;
GLuint fbTex;
void initFrameBuffer();
Shader* quadShader;
#pragma endregion


#pragma region lights
//glm::vec3 lightColor = glm::vec3(0.8f, 0.8f, 0.2f);
//glm::vec3 lightCloseColor = glm::vec3(0.3f, 0.3f, 0.f);
//DirLight dirLight(glm::vec3(0.05f, 0.05f, 0.05f), glm::vec3(0.4f, 0.4f, 0.4f), glm::vec3(0.5f, 0.5f, 0.5f), glm::vec3(-0.2f, -1.0f, -0.2f));
DirLight dirLight(glm::vec3(-1.0f, -1.0f, -1.0f), glm::vec3(0.3f), glm::vec3(1.2f), glm::vec3(1.0f));

//glm::vec3 tryPos = glm::vec3(-1.95f, 2.3f, 2.51f);
//glm::vec3 tryScale = glm::vec3(0.5f, 0.5f, .1f);
#pragma endregion


#pragma region shadow
Shader* depthShader;
Shader* debugDepthQuad;

Shadow* shadow;
int debugLayer;
bool showQuad = false;
void initShadow();
#pragma endregion

#pragma region camera
// camera
Camera camera(glm::vec3(-6.23362f, 7.35175f, 54.5771f), glm::vec3(0.0f, 1.0f, 0.0f), -117.5f, -12.0);
const float CAMERA_NEAR = 0.1f;
const float CAMERA_FAR = 1000.0f;

std::vector<CameraParams> cameraConfig = { 
    {glm::vec3(-6.23362f, 7.35175f, 54.5771f), -117.5f, -12.0f},
    {glm::vec3(-1.13798, 5.942, 44.4472), -117.5f, -12.f},
    {glm::vec3(-0.0199825, 6.29115, 45.2214), -145.3f, -14.4f},
    {glm::vec3(8.59052, 6.29115, 32.7863), -145.3f, -14.4f},
    {glm::vec3(9.58896, 7.69788, 37.5243), -96.7f, -14.0f},
    {glm::vec3(3.51082, 7.69788, 32.3419), -39.1f, -14.8f},
    {glm::vec3(3.51082, 7.69788, 32.3419), -7.69997f, -14.6f},
    {glm::vec3(7.9009, 7.69788, 33.4396), -83.8999f, -10.2f},
    {glm::vec3(16.1728, 6.9645, 28.1266), -131.9f, -12.8f},
    {glm::vec3(16.1728, 6.9645, 28.1266), -137.1f, 35.8f},  // fireIdx = 9
    {glm::vec3(16.1728, 6.9645, 28.1266), -137.1f, 35.8f},
    {glm::vec3(16.1728, 6.9645, 28.1266), -139.1f, -37.f},
    {glm::vec3(16.1728, 6.9645, 28.1266), -139.1f, -37.f},
    {glm::vec3(17.1856, 8.58474, 27.1655), -127.9f, -9.60001f},
    {glm::vec3(16.401, 4.9162, 1.70374), -114.1f, -6.60001f},
    {glm::vec3(11.6809, 2.53701, -18.711), -114.9f, -3.60001f},
    {glm::vec3(11.6809, 2.53701, -18.711), -6.89954f, -2.20001f},
    {glm::vec3(11.6809, 2.53701, -18.711), 66.9005f, 6.19999f},
    {glm::vec3(11.6809, 2.53701, -18.711), 58.1005f, 42.6f}
};
const int fireIdx = 9;
float curTime = 0.0f;
const float intervals = 10.0f;
float lastX = SCR_WIDTH / 2.0f;
float lastY = SCR_HEIGHT / 2.0f;
bool cameraRot = false;
bool firstMove = true;
CameraParams lerp(CameraParams p0, CameraParams p1, float ratio);
void roam(float deltaTime);
#pragma endregion

Skybox skybox;
Launcher launcher;


int main()
{
    srand(time(NULL));
    // glfw: initialize and configure
    // ------------------------------
    glfwInit();
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 6);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_SAMPLES, 4);

#ifdef __APPLE__
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
#endif



    // glfw window creation
    // --------------------
    //GLFWwindow* window = glfwCreateWindow(SCR_WIDTH, SCR_HEIGHT, "main", NULL, NULL);
    GLFWwindow* window = glfwCreateWindow(SCR_WIDTH, SCR_HEIGHT, "main", glfwGetPrimaryMonitor(), NULL);
    if (window == NULL)
    {
        std::cout << "Failed to create GLFW window" << std::endl;
        glfwTerminate();
        return -1;
    }
    glfwMakeContextCurrent(window);
    glfwSetFramebufferSizeCallback(window, framebuffer_size_callback);
    glfwSetCursorPosCallback(window, mouse_callback);
    glfwSetMouseButtonCallback(window, mouse_button_callback);
    glfwSetScrollCallback(window, scroll_callback);
    glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);

    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
    {
        std::cout << "Failed to initialize GLAD" << std::endl;
        return -1;
    }
    stbi_set_flip_vertically_on_load(true);

    projection = glm::perspective(glm::radians(camera.Zoom), (float)SCR_WIDTH / (float)SCR_HEIGHT, CAMERA_NEAR, CAMERA_FAR);
    glClearColor(0.5f, 0.5f, 0.5f, 1.0f);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    //glEnable(GL_CULL_FACE);
    //glFrontFace(GL_CCW);
    //glCullFace(GL_BACK);
    //glEnable(GL_MULTISAMPLE);
    glEnable(GL_CULL_FACE);

    cubeVAO = Utils::InitCube();
    quadVAO = Utils::InitQuad();
    
    std::cout << "Program start..." << std::endl;
    std::cout << "Shader path: " << SHADER << std::endl;
    std::cout << "Texture path: " << TEX << std::endl;
    std::cout << "Object path: " << OBJ << std::endl;

    skybox.Init(&camera, projection);
    launcher.Init(&camera, projection);
    std::cout << "Loading shader: " << SHADER"particle_compute.cs" << std::endl;
    initFrameBuffer();
    initModels();
    initShadow();

    //camera.Auto = true;

    // render loop
    // -----------
    while (!glfwWindowShouldClose(window))
    {
        float currentFrame = glfwGetTime();
        deltaTime = currentFrame - lastFrame;
        lastFrame = currentFrame;
        processInput(window);

        fluidShader->use();
        fluidShader->setInt("water_effect", fluidEffect);
        fluidShader->setFloat("time", glfwGetTime());
        if (camera.Position.y < 0.0f)
            fluidShader->setFloat("fresnel_base", 0.0f);
        else
            fluidShader->setFloat("fresnel_base", 0.15f);
        if (camera.Pitch < -45.0f || camera.Pitch > 45.0f) 
            fluidShader->setBool("inv_tex_reflection", true);
        else 
            fluidShader->setBool("inv_tex_reflection", false);

        glm::vec4 clippingBounds = glm::vec4(-50.0f, 50.0f, -50.0f, 50.0f);
        fluidShader->setVec4("clippingBounds", clippingBounds);

        launcher.Update(deltaTime);
        roam(deltaTime);

        // render to depth framebuffer
        shadow->SetupUBO(-dirLight.direction);
        shadow->Bind();
        glClear(GL_DEPTH_BUFFER_BIT);
        glCullFace(GL_FRONT);  // peter panning
        //renderScene(simpleDepthShader);
        renderModels(camera.GetViewMatrix(), false, false, true);  //dont display water in refraction
        glCullFace(GL_BACK);
        shadow->Unbind();
        
        // render to refraction framebuffer
        fbRefraction->Bind();
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        renderModels(camera.GetViewMatrix(), false, false, false);  //dont display water in refraction
        launcher.Draw(false);
        skybox.Draw(false);
        fbRefraction->Unbind();

        // render to reflection framebuffer
        fbReflection->Bind();
        //glEnable(GL_CLIP_DISTANCE0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        // use the different view matrix which is from under the water
        renderModels(camera.GetReflectionMatrix(2.0f), true, false, false);
        launcher.Draw(true);
        skybox.Draw(true);
        //glDisable(GL_CLIP_DISTANCE0);
        fbReflection->Unbind();
        
        // final framebuffer
        fb->Bind(); //print to the real, color, multisample frambuffer
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        renderModels(camera.GetViewMatrix(), false, true, false);
        launcher.Draw(false);
        skybox.Draw(false);
        fb->Unbind();

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        quadShader->use();
        glActiveTexture(GL_TEXTURE0);
        if (showQuad) {
            glBindTexture(GL_TEXTURE_2D, fbReflection->GetTexture());
            quadShader->setFloat("tex", fbReflection->GetTexture());
        }
        else {
            glBindTexture(GL_TEXTURE_2D, fbTex);
            quadShader->setFloat("tex", fbTex);
        }
        Utils::RenderQuad(quadVAO);

        //glBindTexture(GL_TEXTURE_2D_ARRAY, shadow->GetTexture());
        //debugDepthQuad->use();
        //debugDepthQuad->setInt("layer", debugLayer);
        //glActiveTexture(GL_TEXTURE0);
        //glBindTexture(GL_TEXTURE_2D_ARRAY, shadow->GetTexture());
        //if (showQuad)
        //{
        //    Utils::RenderQuad(quadVAO);
        //}

        // -------------------------------------------------------------------------------
        glfwSwapBuffers(window);
        glfwPollEvents();

        //cout << "{glm::vec3(" << camera.Position.x << ", " << camera.Position.y << ", " << camera.Position.z << "), " << camera.Yaw << "f, " << camera.Pitch << "f}," << endl;
    }

    //glfwTerminate();
    //std::cout << "Press any key to exit..." << std::endl;
    //std::cin.get();
 

    return 0;
}

void initModels()
{
    group = new Model(OBJ"LAKE.obj");
    terrain = &(group->meshes[0]);
    fluid = &(group->meshes[1]);

    forest = new Model(OBJ"Trees.obj");
    houses = new Model(OBJ"House.obj");

    staticShader = new Shader(SHADER"model.vs", SHADER"model.fs");

    fluidShader = new Shader(SHADER"fluid.vs", SHADER"fluid.fs");
    fluidShader->use();

    fluid->SetRefractionTex(*fluidShader, fbRefraction->GetTexture());
    fluid->SetReflectionTex(*fluidShader, fbReflection->GetTexture());
    fluid->SetDiffuseTex(*fluidShader);
}

void initFrameBuffer()
{
    fbRefraction = new Framebuffer();
    fbRefraction->Init(SCR_WIDTH, SCR_HEIGHT, false);
    fbReflection = new Framebuffer();
    fbReflection->Init(SCR_WIDTH, SCR_HEIGHT, false);

    fb = new Framebuffer();
    fbTex = fb->Init(SCR_WIDTH, SCR_HEIGHT, true);
    quadShader = new Shader(SHADER"quad_screen.vs", SHADER"quad_screen.fs");
}

void initShadow()
{
    depthShader = new Shader(SHADER"shadow_mapping_depth.vs", SHADER"shadow_mapping_depth.fs", SHADER"shadow_mapping_depth.gs");
    shadow = new Shadow();
    shadow->Init(SCR_WIDTH, SCR_HEIGHT, CAMERA_NEAR, CAMERA_FAR, &camera);

    debugDepthQuad = new Shader(SHADER"debug_quad.vs", SHADER"debug_quad_depth.fs");

    debugDepthQuad->use();
    debugDepthQuad->setInt("depthMap", 0);
}

void renderModels(glm::mat4 view, bool isInvCam, bool isDrawFluid, bool isDrawDepth)
{
    glm::mat4 model = glm::mat4(1.0f);
    model = glm::mat4(1.0f);
    //model = glm::translate(model, furnacePos);
    model = glm::scale(model, glm::vec3(5.0f, 5.f, 5.f));

    Shader* shader = isDrawDepth ? depthShader : staticShader;
    shader->use();

    if (isDrawDepth) {
        shader->setMat4("model", model);
    }
    else {
        dirLight.SetupInShader(*shader, "dirLight");
        shader->setMat4("projection", projection);
        shader->setMat4("model", model);
        shader->setMat4("view", view);
        shader->setVec3("viewPos", camera.Position.x, isInvCam ? -camera.Position.y : camera.Position.y, camera.Position.z);
        shader->setFloat("farPlane", shadow->GetCameraFarPlane());
        shader->setInt("cascadeCount", shadow->GetShadowCascadeLevels().size());
        
        shader->setBool("hideBelow", isInvCam);

        for (size_t i = 0; i < shadow->GetShadowCascadeLevels().size(); ++i)
        {
            shader->setFloat("cascadePlaneDistances[" + std::to_string(i) + "]", shadow->GetShadowCascadeLevels()[i]);
        }
        //glActiveTexture(GL_TEXTURE0 + terrain->textures.size());
        //glBindTexture(GL_TEXTURE_2D_ARRAY, shadow->GetTexture());
        glActiveTexture(GL_TEXTURE2);
        glBindTexture(GL_TEXTURE_2D_ARRAY, shadow->GetTexture());
        shader->setInt("shadowMap", 2);
    }

    terrain->Draw(*shader, !isDrawDepth);
    forest->Draw(*shader, !isDrawDepth);
    houses->Draw(*shader, !isDrawDepth);

    if (isDrawFluid) {
        //glm::mat4 fluidScale = glm::scale(glm::mat4(1.0f), glm::vec3(3.0f, 1.0f, 3.0f));
		glm::mat4 fluidScale = model;
        fluidShader->use();
        dirLight.SetupInShader(*fluidShader, "dirLight");
        //fluidShader->setInt("water_effect", fluidEffect);
        //fluidShader->setFloat("time", glfwGetTime());
        fluidShader->setMat4("projection", projection);
        fluidShader->setMat4("model", fluidScale);
        fluidShader->setMat4("view", view);
        fluidShader->setVec3("viewPos", camera.Position.x, camera.Position.y, camera.Position.z);
        glDisable(GL_CULL_FACE);
        fluid->Draw();
        glEnable(GL_CULL_FACE);
    }
}

CameraParams lerp(CameraParams p0, CameraParams p1, float ratio) {
    assert(ratio >= 0.0f && ratio <= 1.0f);
    return { p0.position + (p1.position - p0.position) * ratio, p0.yaw + (p1.yaw - p0.yaw) * ratio, p0.pitch + (p1.pitch - p0.pitch) * ratio };
}

void roam(float deltaTime) {
    if (camera.Auto == false) return;
    if (curTime > intervals * (cameraConfig.size() - 1)) {
        curTime = false;
        camera.Auto = false;
        return;
    }
    int sId = curTime / intervals;
    int eId = sId == cameraConfig.size() - 1 ? sId : sId + 1;
    if (sId == fireIdx) launcher.Launch = true;
    float ratio = (curTime - sId * intervals) / intervals;
    CameraParams param = lerp(cameraConfig[sId], cameraConfig[eId], ratio);
    camera.Apply(param);
    curTime += deltaTime;
}

// process all input: query GLFW whether relevant keys are pressed/released this frame and react accordingly
// ---------------------------------------------------------------------------------------------------------
void processInput(GLFWwindow* window)
{
    if (glfwGetKey(window, GLFW_KEY_ESCAPE) == GLFW_PRESS)
        glfwSetWindowShouldClose(window, true);

    if (glfwGetKey(window, GLFW_KEY_W) == GLFW_PRESS)
        camera.ProcessKeyboard(FORWARD, deltaTime);
    if (glfwGetKey(window, GLFW_KEY_S) == GLFW_PRESS)
        camera.ProcessKeyboard(BACKWARD, deltaTime);
    if (glfwGetKey(window, GLFW_KEY_A) == GLFW_PRESS)
        camera.ProcessKeyboard(LEFT, deltaTime);
    if (glfwGetKey(window, GLFW_KEY_D) == GLFW_PRESS)
        camera.ProcessKeyboard(RIGHT, deltaTime);
    if (glfwGetKey(window, GLFW_KEY_LEFT_SHIFT) == GLFW_PRESS)
        camera.ProcessKeyboard(UP, deltaTime);
    if (glfwGetKey(window, GLFW_KEY_LEFT_CONTROL) == GLFW_PRESS)
        camera.ProcessKeyboard(DOWN, deltaTime);

    static int spacePress = GLFW_RELEASE;
    if (glfwGetKey(window, GLFW_KEY_SPACE) == GLFW_RELEASE && spacePress == GLFW_PRESS)
    {
        camera.Auto = !camera.Auto;
    }
    spacePress = glfwGetKey(window, GLFW_KEY_SPACE);

    static int fPress = GLFW_RELEASE;
    if (glfwGetKey(window, GLFW_KEY_F) == GLFW_RELEASE && fPress == GLFW_PRESS)
    {
        launcher.Launch = !launcher.Launch;
    }
    fPress = glfwGetKey(window, GLFW_KEY_F);

    //if (glfwGetKey(window, GLFW_KEY_SPACE) == GLFW_PRESS) {
    //    keyPress = true;
    //}
    //if (glfwGetKey(window, GLFW_KEY_SPACE) == GLFW_RELEASE && keyPress) {
    //    keyPress = false;
    //}
    static int dPress = GLFW_RELEASE;
    if (glfwGetKey(window, GLFW_KEY_B) == GLFW_RELEASE && dPress == GLFW_PRESS)
    {
        showQuad = !showQuad;
    }
    dPress = glfwGetKey(window, GLFW_KEY_B);
    //static int plusPress = GLFW_RELEASE;
    //if (glfwGetKey(window, GLFW_KEY_N) == GLFW_RELEASE && plusPress == GLFW_PRESS)
    //{
    //    debugLayer++;
    //    if (debugLayer > shadow->GetShadowCascadeLevels().size())
    //    {
    //        debugLayer = 0;
    //    }
    //}
    //plusPress = glfwGetKey(window, GLFW_KEY_N);


    //if (glfwGetKey(window, GLFW_KEY_I) == GLFW_PRESS)
    //    tryPos.x +=  0.001;
    //if (glfwGetKey(window, GLFW_KEY_K) == GLFW_PRESS)
    //    tryPos.x -= 0.001;
    //if (glfwGetKey(window, GLFW_KEY_U) == GLFW_PRESS)
    //    tryPos.y -= 0.001;
    //if (glfwGetKey(window, GLFW_KEY_O) == GLFW_PRESS)
    //    tryPos.y += 0.001;
    //if (glfwGetKey(window, GLFW_KEY_J) == GLFW_PRESS)
    //    tryPos.z -= 0.001;
    //if (glfwGetKey(window, GLFW_KEY_L) == GLFW_PRESS)
    //    tryPos.z += 0.001;
    //if (glfwGetKey(window, GLFW_KEY_N) == GLFW_PRESS)
    //    tryScale -= glm::vec3(0.001);
    //if (glfwGetKey(window, GLFW_KEY_M) == GLFW_PRESS)
    //    tryScale += glm::vec3(0.001);
    //cout << "omgomg tryPos = " << tryPos.x << ", " << tryPos.y << ", " << tryPos.z << endl;
    //cout << "omgomg tryScale = " << tryScale.x << ", " << tryScale.y << ", " << tryScale.z << endl;

}

void mouse_callback(GLFWwindow* window, double xpos, double ypos)
{
    if (!cameraRot) return;
    if (firstMove)
    {
        lastX = xpos;
        lastY = ypos;
        firstMove = false;
    }

    float xoffset = xpos - lastX;
    float yoffset = lastY - ypos; // reversed since y-coordinates go from bottom to top

    lastX = xpos;
    lastY = ypos;

    camera.ProcessMouseMovement(xoffset, yoffset);
}

void mouse_button_callback(GLFWwindow* window, int button, int action, int mods)
{
    if (action == GLFW_PRESS && button == GLFW_MOUSE_BUTTON_LEFT) {
        cameraRot = true;
    }
    if (action == GLFW_RELEASE && button == GLFW_MOUSE_BUTTON_LEFT) {
        cameraRot = false;
        firstMove = true;
    }

}

void scroll_callback(GLFWwindow* window, double xoffset, double yoffset)
{
    camera.ProcessMouseScroll(yoffset);
}



// ---------------------------------------------------------------------------------------------
void framebuffer_size_callback(GLFWwindow* window, int width, int height)
{
    glViewport(0, 0, width, height);
}
