#ifndef __MESH_HEADER__
#define __MESH_HEADER__

#include "shader.h"

#include <vector>
#include <string>

#include <glm/glm.hpp>

// Only for mesh without texture
struct Material
{
	glm::vec3 sAmbient;
	glm::vec3 sDiffuse;
	glm::vec3 sSpecular;
	float sShininess;
};

struct Vertex
{
	glm::vec3 sPosition;
	glm::vec3 sNormal;
	glm::vec2 sTexCoords;
};

struct Texture
{
	unsigned int sId;
	std::string sType;
	std::string sPath;
};

// Mesh Definition
class Mesh
{
public:
	/* Mesh Data */
	std::vector<Vertex> vertices;
	std::vector<unsigned int> indices;
	std::vector<Texture> textures;
	// Only used when the mesh doesn't have texture
	Material m_material;

	Mesh(std::vector<Vertex> vertices, std::vector<unsigned int> indices, std::vector<Texture> textures)
	{
		this->vertices = vertices;
		this->indices = indices;
		this->textures = textures;
		//std::cout << "omgomg texturesize = " << this->textures.size() << std::endl;
		m_material.sAmbient = glm::vec3(0.2f, 0.2f, 0.2f);
		m_material.sDiffuse = glm::vec3(0.8f, 0.8f, 0.8f);
		m_material.sSpecular = glm::vec3(1.0f, 1.0f, 1.0f);
		m_material.sShininess = 32.0f;

		setupMesh();
	}
	~Mesh(){}

	void SetRefractionTex(Shader& shader, GLuint tex) {
		shader.use();
		m_refractionTex = tex;
		shader.setInt("tex_refraction", 0);
		//m_refractionTex = m_textures[0].sId;
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, m_refractionTex);
		glBindTexture(GL_TEXTURE_2D, 0);
	}

	void SetReflectionTex(Shader& shader, GLuint tex) {
		shader.use();
		m_reflectionTex = tex;
		shader.setInt("tex_reflection", 1);
		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, m_reflectionTex);
		glBindTexture(GL_TEXTURE_2D, 0);
	}

	void SetDiffuseTex(Shader& shader) {
		shader.use();
		shader.setInt("tex_diffuse", 2);
		glActiveTexture(GL_TEXTURE2);
		glBindTexture(GL_TEXTURE_2D, textures[0].sId);
		glBindTexture(GL_TEXTURE_2D, 0);
	}

	void Draw() {
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, m_refractionTex);
		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, m_reflectionTex);
		glActiveTexture(GL_TEXTURE2);
		glBindTexture(GL_TEXTURE_2D, textures[0].sId);
		glBindVertexArray(m_VAO);
		glDrawElements(GL_TRIANGLES, indices.size(), GL_UNSIGNED_INT, 0);
		glBindVertexArray(0);
	}

	void Draw(Shader& shader, bool bindTexture) 
	{
		unsigned int diffuseNr = 1;
		unsigned int specularNr = 1;
		if (bindTexture) {
			for (unsigned int i = 0; i < textures.size(); ++i)
			{
				glActiveTexture(GL_TEXTURE0 + i);

				std::stringstream ss;
				std::string number;
				std::string name = textures[i].sType;

				if (name == "texture_diffuse")
					ss << diffuseNr++;
				else if (name == "texture_specular")
					ss << specularNr++;
				number = ss.str();

				shader.setInt(("material." + name + number).c_str(), i);
				glBindTexture(GL_TEXTURE_2D, textures[i].sId);
			}
			if (textures.size() == 0)
			{
				shader.setFloat("material.shininess", m_material.sShininess);
				shader.setVec3("material.ambient", m_material.sAmbient.x, m_material.sAmbient.y, m_material.sAmbient.z);;
				shader.setVec3("material.diffuse", m_material.sDiffuse.x, m_material.sDiffuse.y, m_material.sDiffuse.z);
				shader.setVec3("material.specular", m_material.sSpecular.x, m_material.sSpecular.y, m_material.sSpecular.z);
				glBindTexture(GL_TEXTURE_2D, 0);
			}
		}

		// draw mesh
		glBindVertexArray(m_VAO);
		glDrawElements(GL_TRIANGLES, indices.size(), GL_UNSIGNED_INT, 0);
		glBindVertexArray(0);
	}

private:
	/* Render data */
	unsigned int m_VAO, m_VBO, m_EBO;

	GLuint m_reflectionTex, m_refractionTex;

	void setupMesh()
	{
		glGenVertexArrays(1, &m_VAO);
		glGenBuffers(1, &m_VBO);
		glGenBuffers(1, &m_EBO);

		glBindVertexArray(m_VAO);

		glBindBuffer(GL_ARRAY_BUFFER, m_VBO);
		glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(Vertex), &vertices[0], GL_STATIC_DRAW);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_EBO);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(unsigned int), &indices[0], GL_STATIC_DRAW);

		// vertex positions
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)0);
		// vertex normals
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, sNormal));
		// vertex texture coords
		glEnableVertexAttribArray(2);
		glVertexAttribPointer(2, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, sTexCoords));

		glBindVertexArray(0);
	}
};

// Skeleton Definition
/*struct Skeleton
{
	std::vector<unsigned int> sChildrenSkeletonsIndices;
	int sParentSkeletonIndex; // -1 == root
	glm::vec3 sGlobalTranslation;
	std::string sJointName;
};*/


#endif
